cimport pdc.cpdc as cpdc
from pdc.cpdc cimport pdc_var_type_t, int64_t, uint64_t, int16_t, int8_t
from typing import TypeVar, Generic, NewType, Union
from enum import Enum
import os
from cpython.mem cimport PyMem_Malloc as malloc, PyMem_Free as free
import numpy as np
from abc import ABC, abstractmethod
from weakref import finalize
import shutil
from subprocess import Popen, PIPE
import ast
import numpy.typing as npt

int32 = NewType('int32', int)
uint32 = NewType('uint32', int)
uint64 = NewType('uint64', int)
pdcid = NewType('pdcid', int)

cdef void *malloc_or_memerr(size_t size):
    cdef void *ptr = malloc(size)
    if not ptr:
        raise MemoryError()
    return ptr

def _free_from_int(ptr: int):
    cdef size_t cptr
    if ptr:
        cptr = ptr
        free(<void *> cptr)

ctrace_id_map = {}
term_color_map = {
    'red': '\033[31m',
    'green': '\033[32m',
    'yellow': '\033[33m',
    'blue': '\033[34m',
    'magenta': '\033[35m',
    'turquoise': '\033[36m'
}
term_color_end = '\033[0m'
last_color = None
do_ctrace = False
do_ctrace_print = False
ctrace_file = None
def format_id(id):
    global last_color
    if id in ctrace_id_map:
        color = ctrace_id_map[id]
    elif last_color is None:
        color = 'red'
    else:
        colors = list(term_color_map.keys())
        color = colors[(colors.index(last_color) + 1) % len(colors)]
    
    ctrace_id_map[id] = color
    last_color = color
    return term_color_map[color] + str(id) + term_color_end
def format_arg(arg):
    if isinstance(arg, int) and arg > 100000:
        return format_id(arg)
    else:
        return str(arg)
def ctrace(name, rtn, *args):
    global ctrace_file
    if not do_ctrace:
        return
    
    if do_ctrace_print:
        print(f'{name}({", ".join([format_arg(i) for i in args])}) -> {format_arg(rtn)}')
    
    #delete old ctrace file
    if ctrace_file is None:
        if os.path.exists('./ctrace.txt'):
            os.remove('./ctrace.txt')
        ctrace_file = open('./ctrace.txt', 'w')
    
    print(f'{name}({", ".join([format_arg(i) for i in args])}) -> {format_arg(rtn)}', file=ctrace_file, flush=True)

class PDCError(Exception):
    '''
    A general error type that indicates an error from the underlying c pdc api.
    Don't count on getting this error most of the time.  Usually, the server/client segfaults instead.
    '''
    pass

_is_open = False

def _get_pdcid() -> pdcid:
    global pdc_id
    return pdc_id

def init(name:str="PDC"):
    '''
    Initialize PDC.
    This call will block and eventually terminate the program if the PDC server is not running in the same directory as the program.
    This must be called before any other PDC method.
    '''
    global pdc_id, _is_open
    name_bytes = name.encode('utf-8')
    pdc_id = cpdc.PDCinit(name_bytes)
    ctrace('init', pdc_id, name)
    if pdc_id == 0:
        raise PDCError('Could not initialize PDC')
    
    #hack to ensure close runs after all finalizers
    #See: https://stackoverflow.com/questions/72622803/how-do-i-make-an-exit-handler-that-runs-after-all-weakref-finalizers-run
    finalize(_close, _close, pdc_id)
    _is_open = True

def _close(pdc_id):
    global _is_open
    if not _is_open:
        return
    err = cpdc.PDCclose(pdc_id)
    ctrace('close', err, pdc_id)
    if err != 0:
        raise PDCError('Could not close PDC')
    _is_open = False

def _get_pdcid():
    global pdc_id
    return pdc_id

def ready() -> bool:
    '''
    True if :func:`init` has been called, False otherwise
    '''
    return _is_open

class Type(Enum):
    '''
    All of the data types that PDC objects support
    '''

    INT      = pdc_var_type_t.PDC_INT
    FLOAT    = pdc_var_type_t.PDC_FLOAT
    DOUBLE   = pdc_var_type_t.PDC_DOUBLE
    #removed for being identical to INT8:
    #CHAR     = pdc_var_type_t.PDC_CHAR
    UINT     = pdc_var_type_t.PDC_UINT
    INT64    = pdc_var_type_t.PDC_INT64
    UINT64   = pdc_var_type_t.PDC_UINT64
    INT16    = pdc_var_type_t.PDC_INT16
    INT8     = pdc_var_type_t.PDC_INT8

    def as_numpy_type(self):
        '''
        Returns the numpy type corresponding to this PDC type
        '''
        if self == Type.INT:
            return np.dtype('=i4')
        elif self == Type.UINT:
            return np.dtype('=u4')
        elif self == Type.FLOAT:
            return np.dtype('=f')
        elif self == Type.DOUBLE:
            return np.dtype('=d')
        elif self == Type.INT64:
            return np.dtype('=i8')
        elif self == Type.UINT64:
            return np.dtype('=u8')
        elif self == Type.INT16:
            return np.dtype('=h')
        elif self == Type.INT8:
            return np.dtype('=b')
        else:
            raise ValueError('Unknown PDC type')
    
    @staticmethod
    def from_numpy_type(dtype):
        '''
        Returns the PDC type corresponding to this numpy type

        :param dtype: The numpy type
        :raises ValueError: If the numpy type has no corresponding PDC type
        '''
        if dtype == np.dtype('=i4'):
            return Type.INT
        elif dtype == np.dtype('=u4'):
            return Type.UINT
        elif dtype == np.dtype('=f'):
            return Type.FLOAT
        elif dtype == np.dtype('=d'):
            return Type.DOUBLE
        elif dtype == np.dtype('=c'):
            return Type.INT8
        elif dtype == np.dtype('=i8'):
            return Type.INT64
        elif dtype == np.dtype('=u8'):
            return Type.UINT64
        elif dtype == np.dtype('=h'):
            return Type.INT16
        elif dtype == np.dtype('=b'):
            return Type.INT8
        else:
            raise ValueError(f'Unsupported numpy type: {dtype}')
    
    '''
    currently unused
    def validate(self, value:object):
        cdef int int_
        cdef unsigned int uint_
        cdef float float_
        cdef double double_
        cdef char char_
        cdef int64_t int64_
        cdef uint64_t uint64_ 
        cdef int16_t int16_
        cdef int8_t int8_
        if self in (Type.FLOAT, Type.DOUBLE):
            checktype(value, 'value', float)
        else:
            checktype(value, 'value', float)
        try:
            if self == Type.INT:
                int_ = value
            elif self == Type.UINT:
                uint_ = value
            elif self == Type.FLOAT:
                float_ = value
            elif self == Type.DOUBLE:
                double_ = value
            elif self == Type.CHAR:
                char_ = value
            elif self == Type.INT64:
                int64_ = value
            elif self == Type.UINT64:
                uint64_ = value
            elif self == Type.INT16:
                int16_ = value
            elif self == Type.INT8:
                int8_ = value
            else:
                raise ValueError(f'Unknown PDC type: {self.name}')
        except OverflowError:
            raise OverflowError(f'value {value} is not within the acceptable range for {self.name}') from None
    '''

class KVTags(ABC):
    '''
    An object used to manipulate object and container tags.
    Supports the following operations:

    =================== ==============================
    operation           effect
    =================== ==============================
    ``value = tags[x]`` get the value of the tag ``x``
    ``tags[x] = value`` set the value of the tag ``x``
    ``del tags[x]``     delete the tag ``x``
    =================== ==============================
    '''

    tag_types = (tuple, str, int, float, bool, type(None))
    tag_types_union = Union[tuple, str, int, float, bool, type(None)]

    @classmethod
    def _encode(cls, obj:tag_types_union) -> bytes:
        if isinstance(obj, tuple):
            for i in obj:
                if not isinstance(i, cls.tag_types):
                    raise ValueError('tuples must contain only strings, ints, floats, and tuples')
            if len(obj) == 0:
                return b'()'
            elif len(obj) == 1:
                return b'(' + cls._encode(obj[0]) + b',)'
            else:
                return b'(' + b','.join([cls._encode(i) for i in obj]) + b')'
        elif isinstance(obj, str):
            return repr(obj).encode('utf-8')
        elif isinstance(obj, int):
            return str(obj).encode('utf-8')
        elif isinstance(obj, float):
            return str(obj).encode('utf-8')
        elif isinstance(obj, bool):
            return str(obj).encode('utf-8')
        elif obj is None:
            return b'None'
        else:
            raise TypeError(f'Unsupported type of tag: {type(obj)}')
    
    @staticmethod
    def _decode(data:bytes) -> tag_types_union:
        return ast.literal_eval(data.decode('utf-8'))
    
    @abstractmethod
    def set(self, key:bytes, value:bytes):
        '''
        :meta private:
        '''
        pass
    
    @abstractmethod
    def get(self, key:bytes) -> bytes:
        '''
        :meta private:
        '''
        pass
    
    @abstractmethod
    def delete(self, key:bytes):
        '''
        :meta private:
        '''
        pass

    def __getitem__(self, name:str):
        checktype(name, 'tag name', str)
        return type(self)._decode(self.get(name.encode('utf-8')))
    
    def __setitem__(self, name:str, value:tag_types_union):
        checktype(name, 'tag name', str)
        self.set(name.encode('utf-8'), type(self)._encode(value))
    
    def __delitem__(self, name:str):
        checktype(name, 'tag name', str)
        self.delete(name.encode('utf-8'))

class ServerContext:
    '''
    A context manager that starts and stops a single PDC server instance on enter and exit.
    This is intended for testing.
    
    Usage::
    
        with pdc.ServerContext():
            pdc.init()
            ...
    
    :raises FileNotFoundError: if pdc_server.exe is not on PATH
    '''
    def __enter__(self):
        path = shutil.which('pdc_server.exe')
        if not path:
            raise FileNotFoundError('pdc_server.exe not on PATH')
        self.popen = Popen([path], stdout=PIPE, encoding='ascii', text=True)
        for line in self.popen.stdout:
            if 'Server ready' in line:
                break
        else:
            raise PDCError('Could not start PDC server')
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.popen.kill()
        self.popen.wait()
        return False

def checktype(value, name, expected):
    if not isinstance(value, expected):
        raise TypeError(f'invalid type of {name}: {type(value)}, expected {expected.__name__}')

def checkrange(value, name, expected:str):
    if expected == 'uint32':
        if value < 0 or value > 0xFFFFFFFF:
            raise OverflowError(f'{name} is out of range for uint32: {value}')
    elif expected == 'uint64':
        if value < 0 or value > 0xFFFFFFFFFFFFFFFF:
            raise OverflowError(f'{name} is out of range for uint64: {value}')
    elif expected == 'int32':
        if value < -0x7FFFFFFF or value > 0x7FFFFFFF:
            raise OverflowError(f'{name} is out of range for int32: {value}')
    else:
        raise ValueError(f'invalid value for expected range: {expected}')



from libc.stdint cimport uint32_t, uint64_t, int8_t, int16_t, int64_t

cdef extern from "pdc_public.h":
    ctypedef int                perr_t
    ctypedef uint64_t           pdcid_t
    ctypedef unsigned long long psize_t
    ctypedef bint               pbool_t

    ctypedef int PDC_int_t

    ctypedef enum pdc_lifetime_t:
        PDC_PERSIST
        PDC_TRANSIENT
    ctypedef enum:
        SUCCEED = 0
        FAIL = -1
    ctypedef enum pdc_var_type_t:
        PDC_UNKNOWN  = -1
        PDC_INT      = 0
        PDC_FLOAT    = 1
        PDC_DOUBLE   = 2
        PDC_CHAR     = 3
        PDC_COMPOUND = 4
        PDC_ENUM     = 5
        PDC_ARRAY    = 6
        PDC_UINT     = 7
        PDC_INT64    = 8
        PDC_UINT64   = 9
        PDC_INT16    = 10
        PDC_INT8     = 11
        NCLASSES     = 12
    
    ctypedef struct pdc_histogram_t:
        pdc_var_type_t dtype
        int            nbin
        double         incr
        double *       range
        uint64_t *     bin

cdef extern from "pdc.h":
    pdcid_t PDCinit(const char *pdc_name)
    perr_t PDCclose(pdcid_t pdcid)

cdef extern from "pdc_id_pkg.h":
    ctypedef struct _pdc_id_info:
        pdcid_t           id
        void *            obj_ptr

cdef extern from "pdc_cont.h":
    ctypedef _pdc_id_info cont_handle
    struct pdc_cont_info:
        char *   name
        pdcid_t  local_id
        uint64_t meta_id
    
    pdcid_t PDCcont_create(const char *cont_name, pdcid_t cont_create_prop)
    #pdcid_t PDCcont_create_col(const char *cont_name, pdcid_t cont_prop_id)
    pdcid_t PDCcont_open(const char *cont_name, pdcid_t pdc_id)
    #pdcid_t PDCcont_open_col(const char *cont_name, pdcid_t pdc_id)
    perr_t PDCcont_close(pdcid_t cont_id)
    perr_t PDCcont_persist(pdcid_t cont_id)
    perr_t PDCprop_set_cont_lifetime(pdcid_t cont_create_prop, pdc_lifetime_t cont_lifetime)
    pdcid_t PDCcont_get_id(const char *cont_name, pdcid_t pdc_id)

    #low priority, but straightforward to implement
    cont_handle *PDCcont_iter_start()
    cont_handle *PDCcont_iter_next(cont_handle *chandle)
    pbool_t PDCcont_iter_null(cont_handle *chandle)
    pdc_cont_info *PDCcont_iter_get_info(cont_handle *chandle)
    
    perr_t PDCcont_del(pdcid_t cont_id)
    perr_t PDCcont_put_objids(pdcid_t cont_id, int nobj, pdcid_t *obj_ids)
    perr_t PDCcont_del_objids(pdcid_t cont_id, int nobj, pdcid_t *obj_ids)
    perr_t PDCcont_put_tag(pdcid_t cont_id, char *tag_name, void *tag_value, psize_t value_size)
    perr_t PDCcont_get_tag(pdcid_t cont_id, char *tag_name, void **tag_value, psize_t *value_size)
    perr_t PDCcont_del_tag(pdcid_t cont_id, char *tag_name)

cdef extern from "pdc_cont_pkg.h":
    struct _pdc_cont_info:
        pdc_cont_info * cont_info_pub
        _pdc_cont_prop *cont_pt
    
    _pdc_cont_info *PDC_cont_get_info(pdcid_t cont_id)

cdef extern from "pdc_obj.h":
    ctypedef _pdc_id_info obj_handle
    struct pdc_obj_info:
        char *       name
        pdcid_t      meta_id
        pdcid_t      local_id
        int          server_id
        pdc_obj_prop *obj_pt
    ctypedef enum pdc_access_t:
        PDC_NA = 0
        PDC_READ = 1
        PDC_WRITE = 2

    pdcid_t PDCobj_create(pdcid_t cont_id, const char *obj_name, pdcid_t obj_create_prop)
    pdcid_t PDCobj_open(const char *obj_name, pdcid_t pdc_id)
    perr_t PDCobj_del(pdcid_t obj_id)
    #pdcid_t PDCobj_open_col(const char *obj_name, pdcid_t pdc_id)
    perr_t PDCobj_close(pdcid_t obj_id)
    pdc_obj_info *PDCobj_get_info(pdcid_t obj)
    perr_t PDCprop_set_obj_user_id(pdcid_t obj_prop, uint32_t user_id)
    perr_t PDCprop_set_obj_app_name(pdcid_t obj_prop, char *app_name)
    perr_t PDCprop_set_obj_time_step(pdcid_t obj_prop, uint32_t time_step)    
    perr_t PDCprop_set_obj_dims(pdcid_t obj_prop, PDC_int_t ndim, uint64_t *dims)
    perr_t PDCprop_set_obj_type(pdcid_t obj_prop, pdc_var_type_t type)
    _pdc_obj_info *PDC_obj_get_info(pdcid_t obj_id)

    #low priority, but straightforward to implement:
    obj_handle *PDCobj_iter_start(pdcid_t cont_id)
    pbool_t PDCobj_iter_null(obj_handle *ohandle)
    obj_handle *PDCobj_iter_next(obj_handle *ohandle, pdcid_t cont_id)
    pdc_obj_info *PDCobj_iter_get_info(obj_handle *ohandle)
    pdcid_t PDCobj_put_data(const char *obj_name, void *data, uint64_t size, pdcid_t cont_id)
    perr_t PDCobj_get_data(pdcid_t obj_id, void *data, uint64_t size)
    perr_t PDCobj_del_data(pdcid_t obj_id)
    perr_t PDCobj_put_tag(pdcid_t obj_id, char *tag_name, void *tag_value, psize_t value_size)
    perr_t PDCobj_get_tag(pdcid_t obj_id, char *tag_name, void **tag_value, psize_t *value_size)
    perr_t PDCobj_del_tag(pdcid_t obj_id, char *tag_name)

cdef extern from "pdc_obj_pkg.h":
    cdef struct _pdc_obj_info:
        _pdc_obj_prop *      obj_pt

#low priority:

#cdef extern from "pdc_analysis.h":
#    pdcid_t PDCobj_data_iter_create(pdcid_t obj_id, pdcid_t reg_id)
#    pdcid_t PDCobj_data_block_iterator_create(pdcid_t obj_id, pdcid_t reg_id, int contig_blocks)
#    size_t PDCobj_data_getSliceCount(pdcid_t iter)
#    size_t PDCobj_data_getNextBlock(pdcid_t iter, void **nextBlock, size_t *dims)
#    perr_t PDCobj_analysis_register(char *func, pdcid_t iterIn, pdcid_t iterOut)
#    int PDCiter_get_nextId()

#nothing from "pdc_dt_conv.h", "pdc_hist_pkg.h" included because it is never mentioned in the examples or tests

cdef extern from "pdc_prop.h":
    struct pdc_obj_prop:
        #not used:
        #pdcid_t        obj_prop_id
        size_t         ndim
        uint64_t *     dims
        pdc_var_type_t type
    
    ctypedef enum pdc_prop_type_t:
        PDC_CONT_CREATE = 0
        PDC_OBJ_CREATE
    
    pdcid_t PDCprop_create(pdc_prop_type_t type, pdcid_t pdc_id)
    perr_t PDCprop_close(pdcid_t id)
    #does not copy most properties.  not that useful:
    #pdcid_t PDCprop_obj_dup(pdcid_t prop_id)

cdef extern from "pdc_prop_pkg.h":
    struct _pdc_obj_prop:
        pdc_obj_prop *obj_prop_pub
        uint32_t             user_id
        char *               app_name
        uint32_t             time_step
    
    struct _pdc_cont_prop:
        pdc_lifetime_t     cont_life
        _pdc_class *pdc

    _pdc_obj_prop *PDC_obj_prop_get_info(pdcid_t obj_prop)


cdef extern from "pdc_region.h":
    struct pdc_region_info:
        pdcid_t               local_id
        size_t                ndim
        uint64_t *            offset
        uint64_t *            size
        bint                  mapping
        int                   registered_op
        void *                buf
        size_t                unit
    
    ctypedef enum pdc_transfer_status_t:
        PDC_TRANSFER_STATUS_COMPLETE  = 0
        PDC_TRANSFER_STATUS_PENDING   = 1
        PDC_TRANSFER_STATUS_NOT_FOUND = 2
    
    pdcid_t PDCregion_create(psize_t ndims, uint64_t *offset, uint64_t *size)
    perr_t PDCregion_close(pdcid_t region_id)
    pdcid_t PDCregion_transfer_create(void *buf, pdc_access_t access_type, pdcid_t obj_id, pdcid_t local_reg, pdcid_t remote_reg)
    perr_t PDCregion_transfer_start(pdcid_t transfer_request_id)
    #perr_t PDCregion_transfer_start_all(pdcid_t *transfer_request_id, size_t size)
    perr_t PDCregion_transfer_status(pdcid_t transfer_request_id, pdc_transfer_status_t *completed)
    perr_t PDCregion_transfer_wait(pdcid_t transfer_request_id)
    #perr_t PDCregion_transfer_wait_all(pdcid_t *transfer_request_id, size_t size)
    perr_t PDCregion_transfer_close(pdcid_t transfer_request_id)

cdef extern from "pdc_query.h":
    #cdef enum pdc_prop_name_t:
    #    PDC_OBJ_NAME
    #    PDC_CONT_NAME
    #    PDC_APP_NAME
    #    PDC_USER_ID

    ctypedef enum pdc_query_op_t:
        PDC_OP_NONE = 0
        PDC_GT = 1
        PDC_LT = 2
        PDC_GTE = 3
        PDC_LTE = 4
        PDC_EQ = 5

    ctypedef enum pdc_query_combine_op_t:
        PDC_QUERY_NONE = 0
        PDC_QUERY_AND = 1
        PDC_QUERY_OR = 2
    
    ctypedef struct pdc_selection_t:
        pdcid_t   query_id
        size_t    ndim
        uint64_t  nhits
        uint64_t *coords
        uint64_t  coords_alloc

    ctypedef struct pdc_query_constraint_t:
        pdcid_t          obj_id
        pdc_query_op_t   op
        pdc_var_type_t   type
        double           value
        pdc_histogram_t *hist

        int            is_range
        pdc_query_op_t op2
        double         value2

        void *  storage_region_list_head
        pdcid_t origin_server
        int     n_sent
        int     n_recv

    ctypedef struct pdc_query_t:
        pdc_query_constraint_t *constraint
        pdc_query_t *    left
        pdc_query_t *    right
        pdc_query_combine_op_t  combine_op
        pdc_region_info *region
        void *                  region_constraint
        pdc_selection_t *       sel

    pdc_query_t *PDCquery_create(pdcid_t obj_id, pdc_query_op_t op, pdc_var_type_t type, void *value)
    pdc_query_t *PDCquery_and(pdc_query_t *query1, pdc_query_t *query2)
    pdc_query_t *PDCquery_or(pdc_query_t *query1, pdc_query_t *query2)
    #perr_t PDCobj_prop_query(pdcid_t cont_id, pdc_prop_name_t prop_name, void *prop_value, pdcid_t **out_ids, size_t *n_out)
    perr_t PDCquery_sel_region(pdc_query_t *query, pdc_region_info *obj_region)
    perr_t PDCquery_get_selection(pdc_query_t *query, pdc_selection_t *sel)
    perr_t PDCquery_get_nhits(pdc_query_t *query, uint64_t *n)
    perr_t PDCquery_get_data(pdcid_t obj_id, pdc_selection_t *sel, void *obj_data)

    perr_t PDCquery_get_sel_data(pdc_query_t *query, pdc_selection_t *sel, void *data)
    void PDCselection_free(pdc_selection_t *sel)
    void PDCquery_free(pdc_query_t *query)
    void PDCquery_free_all(pdc_query_t *query) #?
    void PDCquery_print(pdc_query_t *query)

cdef extern from "pdc_prop_pkg.h":
    ctypedef struct pdc_kvtag_t:
        char *   name
        uint32_t size
        void *   value

cdef extern from "pdc_malloc.h":
    void *PDC_free(void *mem)

cdef extern from "pdc_private.h":
    cdef struct _pdc_class:
        char *  name
        pdcid_t local_id


#cdef extern from "pdc_client_connect.h":
#    perr_t PDC_Client_query_kvtag(pdc_kvtag_t *kvtag, int *n_res, uint64_t **pdc_ids)

#low priority:
#cdef extern from "pdc_transform.h":
#    ctypedef enum pdc_obj_transform_t:
#        PDC_TESTING       = 0
#        PDC_FILE_IO       = 1
#        PDC_DATA_MAP      = 2
#        PDC_PRE_ANALYSIS  = 4
#        PDC_POST_ANALYSIS = 8
#    
#    ctypedef enum pdc_data_movement_t:
#        DATA_IN = 1
#        DATA_OUT = 2
#        DATA_RELOCATION = 4
#    
#    #These functions will have to be changed significantly to be able to accept python functions as arguments
#    #They will need to accept a function pointer directly and an arbitrary void* argument to pass to the function
#    perr_t PDCobj_transform_register(char *func, pdcid_t obj_id, int current_state, int next_state, pdc_obj_transform_t op_type, pdc_data_movement_t when) #?
#    perr_t PDCbuf_map_transform_register(char *func, void *buf, pdcid_t src_region_id, pdcid_t dest_object_id, pdcid_t dest_region_id, int current_state, int next_state, pdc_data_movement_t when) #?



from enum import Enum
from typing import Iterable, Optional, Sequence, Union
from weakref import finalize, WeakValueDictionary
import numpy as np
import numpy.typing as npt

from .main import KVTags, checktype, _get_pdcid, PDCError, ctrace, Type
from pdc.cpdc cimport pdc_lifetime_t, pdc_prop_type_t, pdcid_t, _pdc_cont_info, psize_t, uint64_t, obj_handle, pdc_obj_info, cont_handle, pdc_cont_info
from pdc.main cimport malloc_or_memerr
import pdc
cimport pdc.cpdc as cpdc
from cpython.mem cimport PyMem_Free as free
from cpython.bytes cimport PyBytes_FromStringAndSize

cdef _pdc_cont_info *get_info_struct(pdcid_t id):
    cdef _pdc_cont_info *info = cpdc.PDC_cont_get_info(id)
    if info == NULL:
        raise PDCError("Could not get info for object")
    return info

cdef void free_info_struct(_pdc_cont_info *info):
    free(<void *> <size_t> info[0].cont_pt[0].pdc[0].name)
    cpdc.PDC_free(<void *> <size_t> info[0].cont_pt[0].pdc)
    cpdc.PDC_free(<void *> <size_t> info[0].cont_pt)
    #broken:
    #free(<void *> <size_t> info[0].cont_info_pub[0].name)
    #cpdc.PDC_free(<void *> <size_t> info[0].cont_info_pub)
    cpdc.PDC_free(<void *> <size_t> info)

class ContainerKVTags(KVTags):
    def __init__(self, cont):
        self.cont = cont
    
    def delete(self, name:bytes):
        raise NotImplementedError("Cannot delete tag from container: https://github.com/hpc-io/pdc/issues/66")
        #if cpdc.PDCcont_del_tag(self.cont._id, name) != 0:
        #    raise PDCError("tag does not exist or could not delete tag")
    
    def get(self, name:bytes):
        cdef void* value_out
        cdef psize_t size_out
        if cpdc.PDCcont_get_tag(self.cont._id, name, &value_out, &size_out) != 0:
            raise PDCError("tag does not exist or could not get tag")
        rtn = PyBytes_FromStringAndSize(<char *> value_out, size_out)
        return rtn
    
    def set(self, name:bytes, value:bytes):
        if cpdc.PDCcont_put_tag(self.cont._id, name, <char *> value, len(value)) != 0:
            raise PDCError("could not set tag")

def to_c_object_list(objects: Union['pdc.Object', Sequence['pdc.Object']]):
    cdef size_t size
    cdef pdcid_t *ids
    if isinstance(objects, pdc.Object):
        size = 1
        ids = <pdcid_t *> malloc_or_memerr(sizeof(pdcid_t))
        ids[0] = objects._id
    else:
        size = len(objects)
        ids = <pdcid_t *> malloc_or_memerr(sizeof(pdcid_t) * size)
        try:
            for i in range(size):
                ids[i] = objects[i]._id
        finally:
            free(ids)
    return <uint64_t> ids, size

class Container:
    '''
    | Containers can contain one or more objects, and objects can be part of one or more containers.
    | Containers can be used to allow objects to persist, and have no other functionality at the moment.
    | Every object must be created belonging to a container.
    '''

    class Lifetime(Enum):
        '''
        The possible lifetimes of containers.
        '''
        PERSISTENT = pdc_lifetime_t.PDC_PERSIST
        TRANSIENT  = pdc_lifetime_t.PDC_TRANSIENT
    
    containers_by_id = WeakValueDictionary()
    
    def __init__(self, name:str, lifetime:Lifetime=Lifetime.TRANSIENT, *, _id:Optional[int]=None):
        '''
        __init__(self, name:str, lifetime:Lifetime=Lifetime.TRANSIENT)
        :param str name: the name of the container.  Container names should be unique between all processes that use PDC.
        :param Lifetime lifetime: the container's lifetime.
        '''

        cdef pdcid_t id, prop_id
        if _id is not None:
            checktype(_id, 'id', int)
            id = _id
        else:
            checktype(name, 'name', str)
            checktype(lifetime, 'lifetime', type(self).Lifetime)
            
            if name == '':
                raise ValueError("Container name cannot be empty")

            prop_id = cpdc.PDCprop_create(pdc_prop_type_t.PDC_CONT_CREATE, _get_pdcid())
            ctrace('prop_create', prop_id, 'PDC_CONT_CREATE', _get_pdcid())
            if prop_id == 0:
                raise PDCError('Failed to create container property')
            
            if lifetime != type(self).Lifetime.PERSISTENT:
                rtn = cpdc.PDCprop_set_cont_lifetime(prop_id, lifetime.value)
                ctrace('prop_set_cont_lifetime', rtn, prop_id, lifetime)
                if rtn != 0:
                    raise PDCError('Failed to set container lifetime')
            
            id = cpdc.PDCcont_create(name.encode('utf-8'), prop_id)
            ctrace('cont_create', id, name.encode('utf-8'), prop_id)
            if id == 0:
                raise PDCError('Failed to create container')
            
            rtn = cpdc.PDCprop_close(prop_id)
            ctrace('prop_close', rtn, prop_id)
            if rtn != 0:
                raise PDCError('Failed to close container property')
        
        finalize(self, self._finalize, id)
        self._id = id
        type(self).containers_by_id[id] = self

    @staticmethod
    def _finalize(pdcid_t id):
        rtn = cpdc.PDCcont_close(id)
        ctrace('cont_close', rtn, id)
        if rtn != 0:
            raise PDCError('Failed to close container')
    
    @classmethod
    def _fromid(cls, pdcid_t id):
        if id in cls.containers_by_id:
            return cls.containers_by_id[id]
        else:
            return cls(None, None, _id=id)

    @classmethod
    def get(cls, name:str) -> 'Container':
        '''
        Get an existing container by name.
        If you try to get a container that does not exist, the server segfaults and the call blocks indeinitely.

        :param str name: the name of the container.
        :return: The container.
        '''
        cdef pdcid_t id = cpdc.PDCcont_open(name.encode('utf-8'), _get_pdcid())
        if id == 0:
            raise PDCError('Container not found or failed to open container')
        return cls._fromid(id)
    
    def persist(self) -> None:
        '''
        Make this container persistent if it wasn't already.
        '''
        rtn = cpdc.PDCcont_persist(self._id)
        ctrace('cont_persist', rtn, self._id)
        if rtn != 0:
            raise PDCError('Failed to make container persistent')
    
    @property
    def lifetime(self) -> Lifetime:
        '''
        The lifetime of this container. read-only.
        '''
        cdef _pdc_cont_info *info = get_info_struct(self._id)
        try:
            return type(self).Lifetime(info[0].cont_pt[0].cont_life)
        finally:
            free_info_struct(info)
    
    @property
    def name(self) -> str:
        '''
        The name of this container. read-only
        '''
        cdef _pdc_cont_info *info = get_info_struct(self._id)
        try:
            return info[0].cont_info_pub[0].name.decode('utf-8')
        finally:
            free_info_struct(info)
    
    @property
    def tags(self) -> KVTags:
        '''
        The tags of this container. See :class:`KVTags` for more information.

        NOTE: currently, deleting tags from containers is not supported.
        '''
        return ContainerKVTags(self)
    
    def all_local_objects(self) -> Iterable['pdc.Object']:
        '''
        Get an iterable through all objects in this container that have been retrieved by the client, and haven't yet been garbage collected.
        Deleting an object while this iterator is in use may cause undefined behavior.

        :return: An iterable of Objects
        '''
        cdef obj_handle *handle = cpdc.PDCobj_iter_start(self._id)
        ctrace('obj_iter_start', <uint64_t> handle, self._id)

        cdef pdc_obj_info *info
        cdef pdcid_t id
        
        while True:
            rtn = cpdc.PDCobj_iter_null(handle)
            ctrace('obj_iter_null', bool(rtn), <uint64_t> handle)
            if rtn:
                break
            
            info = cpdc.PDCobj_iter_get_info(handle)
            ctrace('obj_iter_get_info', <uint64_t> info, <uint64_t> handle)
            if info == NULL:
                raise PDCError('Failed to get object info')
            
            id = info[0].local_id
            if id == 0:
                raise PDCError('Object local ID missing')
            
            yield pdc.Object._fromid(id)
            
            handle = cpdc.PDCobj_iter_next(handle, self._id)
            ctrace('obj_iter_next', <uint64_t> handle, <uint64_t> handle, self._id)
    
    def _add_objects(self, objects:Union[Sequence['pdc.Object'], 'pdc.Object']):
        '''
        Add objects to this container

        :param object: A single Object, or a sequence of Objects to add.
        '''
        ptr, size = to_c_object_list(objects)
        cdef pdcid_t *ids = <pdcid_t *> <uint64_t> ptr
        try:
            rtn = cpdc.PDCcont_put_objids(self._id, size, ids)
            ctrace('cont_put_objids', rtn, self._id, size, tuple(o._id for o in objects))
            if rtn != 0:
                raise PDCError('Failed to add objects to container')
        finally:
            free(ids)
    
    #TODO: when this is implemented, add back this warning:
    #        In particular, do not write ``container.remove_objects(container.all_local_objects())``
    def _remove_objects(self, objects:Union[Sequence['pdc.Object'], 'pdc.Object']):
        '''
        Remove objects from this container
        
        :param object: A single Object, or a sequence of Objects to add.
        '''
        ptr, size = to_c_object_list(objects)
        cdef pdcid_t *ids = <pdcid_t *> <uint64_t> ptr
        try:
            rtn = cpdc.PDCcont_del_objids(self._id, size, ids)
            ctrace('cont_del_objids', rtn, self._id, size, tuple(o._id for o in objects))
            if rtn != 0:
                raise PDCError('Failed to add objects to container')
        finally:
            free(ids)
    
    def _delete(self):
        '''
        Delete this container.
        Do not access any methods or properties of this object after calling.

        NOTE: this is currently broken, and doesn't actually delete the container.
        '''
        if cpdc.PDCcont_del(self._id) != 0:
            raise PDCError('Failed to delete container')
    
    def __eq__(self, other):
        return isinstance(other, Container) and self._id == other._id
    
    def __hash__(self):
        return hash(self._id)
    
    def create_object(self, name:str, properties:'Object.Properties'):
        '''
        Create a new object, placed in this container.  To get an existing object, use Object.get() instead

        :param str name: the name of the object.  Object names should be unique between all processes that use PDC, however, passing in a duplicate name will create the object anyway.
        :param ObjectProperties properties: An ObjectProperties describing the properties of the object.
        '''
        return pdc.Object(name, properties, self)
    
    def object_from_array(self, name:str, array:npt.ArrayLike, type:Optional[Type]=None, prop:Optional['Object.Properties']=None) -> 'Object':
        '''
        object_from_array(self, name:str, array, type:Optional[Type]=None, prop:Optional['Object.Properties']=None)
        Create a new object, placed in this container, from an array-like object.
        To get an existing object, use Object.get() instead

        :param str name: the name of the object.  Object names should be unique between all processes that use PDC, however, passing in a duplicate name will create the object anyway.
        :param array: the array to create the object from.  It must be a numpy array, or be convertible to a numpy array via numpy.array().
        :param type: If specified, the resulting object will have this type, if all elements of the array fit in this type.
        :param prop: If specified, the app name, time step, and user id will be copied from this properties object to the resulting object's properties.
        '''
        if type is None:
            array = np.asanyarray(array)
        else:
            array = np.asanyarray(array, dtype=type.as_numpy_type())
        
        dtype = array.dtype
        if not dtype.isnative:
            array = array.astype(dtype.newbyteorder('='))
        
        if type is not None:
            obj_type = type
        else:
            obj_type = pdc.Type.from_numpy_type(dtype)
        
        if prop is None:
            obj_prop = pdc.Object.Properties(type=obj_type, dims=array.shape)
        else:
            obj_prop = prop.copy()
            obj_prop.type = obj_type
            obj_prop.dims = array.shape
        
        obj = self.create_object(name, obj_prop)
        obj.set_data(array).wait()
        return obj

def all_local_containers() -> Iterable[Container]:
    '''
    Get an iterable of all containers that have been retrieved by the client, and haven't yet been garbage collected.
    Deleting a container after this function is called, then using the iterator may cause undefined behavior.

    :return: An iterable of Containers
    '''
    cdef cont_handle *handle = cpdc.PDCcont_iter_start()
    ctrace('cont_iter_start', <uint64_t> handle)

    cdef pdc_cont_info *info
    cdef pdcid_t id
    
    while True:
        rtn = cpdc.PDCcont_iter_null(handle)
        ctrace('cont_iter_null', bool(rtn), <uint64_t> handle)
        if rtn:
            break
        
        info = cpdc.PDCcont_iter_get_info(handle)
        ctrace('cont_iter_get_info', <uint64_t> info, <uint64_t> handle)
        if info == NULL:
            raise PDCError('Failed to get container info')
        
        id = info[0].local_id
        if id == 0:
            raise PDCError('Container local ID missing')
        
        yield Container._fromid(id)
        
        handle = cpdc.PDCcont_iter_next(handle)
        ctrace('cont_iter_next', <uint64_t> handle, <uint64_t> handle)



from functools import singledispatchmethod
import builtins
from pdc.main cimport malloc_or_memerr
from .main import checktype, uint64, checkrange, PDCError, pdcid, ctrace, _free_from_int
from pdc.cpdc cimport uint64_t, pdcid_t
cimport pdc.cpdc as cpdc
from cpython.mem cimport PyMem_Free as free
from typing import Tuple
import copy

#free region info created by _construct_region_info
cdef void free_region_info(pdc_region_info *region_info):
    free(region_info[0].size)
    free(region_info[0].offset)
    free(region_info)

cdef pdc_region_info *construct_region_info(region, dims):
    size_ptr, offset_ptr = region._get_sizes_offsets(dims)
    cdef uint64_t *size_arr = <uint64_t *> <size_t> size_ptr
    cdef uint64_t *offset_arr = <uint64_t *> <size_t> offset_ptr
    
    cdef pdc_region_info *region_info = <pdc_region_info *> malloc_or_memerr(sizeof(pdc_region_info))
    region_info[0].ndim = len(dims)
    region_info[0].offset = offset_arr
    region_info[0].size = size_arr        
    return region_info

class Region:
    '''
    | A region, which specifies the location of a 'block' of data.  
    | Regions are created by slicing the :class:`region` object.
    | Examples:
    | all data of an object: ``region[:]``
    | the first row of a 2 dimensional object, or the first element of a 1 dimensional object: ``region[0]``
    | the first 3 elements of a 1 dimensional object: ``region[:3]``
    | the next 3 elements of a 1 dimensional object: ``region[3:6]``
    | the last 6 columns of the first 2 rows of an object: ``region[:2, 6:]``
    |
    | The slices used to create a region object may not have step values. (``region[1:100:4]`` is invalid)
    '''
    
    @singledispatchmethod
    def __init__(self, slice):
        raise TypeError(f'invalid type of slice: {type(slice)}')
    
    @staticmethod
    def validate_slice(slice:slice):
        '''
        Validates a slice object.

        :meta private:
        '''
        checktype(slice, 'slice', builtins.slice)
        if slice.start is not None:
            checktype(slice.start, 'start value', int)
            checkrange(slice.start, 'start value', 'uint64')
        if slice.stop is not None:
            checktype(slice.stop, 'stop value', int)
            checkrange(slice.stop, 'stop value', 'uint64')
        if slice.step is not None and slice.step != 1:
            raise ValueError('slice step values are not supported')
        if slice.start is not None and slice.stop is not None and slice.start >= slice.stop:
            raise ValueError('start value cannot be greater or equal to than stop value')

    @__init__.register
    def _(self, slice:int):
        cdef uint64_t cslice = slice
        self.slices = (builtins.slice(cslice, cslice+1),)
    
    @__init__.register
    def _(self, slice:builtins.slice):
        Region.validate_slice(slice)
        self.slices = (slice,)
    
    @__init__.register
    def _(self, slice:tuple):
        checktype(slice, 'slice', tuple)
        if len(slice) == 0:
            raise ValueError('region must have at least one dimension')
        
        newslices = []
        for s in slice:
            if isinstance(s, builtins.slice):
                Region.validate_slice(s)
                newslices.append(s)
            else:
                checktype(s, 'slice', int)
                checkrange(s, 'slice', 'uint64')
                newslices.append(builtins.slice(s, s+1))
        
        self.slices = tuple(newslices)
    
    def is_absolute(self):
        '''
        | A region is absolute if the region's slices have both start and stop values.
        | returns True if the region is absolute
        | Examples:
        | ``region[2:3, :9]`` is not absolute
        | ``region[2:3, 8]`` is absolute
        '''
        for s in self.slices:
            if s.start is None or s.stop is None:
                return False
        return True
    
    def get_absolute(self, dims):
        '''
        | Return a new absolute region, using the given dimensions to determine ommitted start and stop values.
        | Examples:
        | ``region[:].get_absolute((3, 4)) -> region[0:3, 0:4]``
        | ``region[:2, 6:].get_absolute((10, 10)) -> region[0:2, 6:10]``
        '''
        newslices = []
        
        for s, d in zip(self.slices, dims):
            start = s.start if s.start is not None else 0

            if s.stop is not None and s.stop > d:
                raise ValueError(f'slice stop value {s.stop} is greater than the dimension {d}')
            stop = s.stop if s.stop is not None else d
            newslices.append(builtins.slice(start, stop))
        
        while len(newslices) < len(dims):
            newslices.append(builtins.slice(0, dims[len(newslices)]))
        
        return Region(tuple(newslices))
    
    def _get_sizes_offsets(self, dims:Tuple[uint64]) -> Tuple[pdcid, Tuple[uint64, ...]]:
        #returns sizes pointer, offsets pointer
        if len(self.slices) > len(dims):
            raise ValueError('region has more dimensions than object')
        newslices = list(self.slices)
        while len(newslices) < len(dims):
            newslices.append(builtins.slice(None))
        
        cdef uint64_t start, stop
        cdef uint64_t *size_arr = <uint64_t *> malloc_or_memerr(sizeof(uint64_t) * len(dims))
        cdef uint64_t *offset_arr = <uint64_t *> malloc_or_memerr(sizeof(uint64_t) * len(dims))
        try:
            for s, d in zip(newslices, dims):
                if s.stop is not None and s.stop > d:
                    raise ValueError(f'slice stop value {s.stop} is greater than the dimension {d}')
            
            for s, d, i in zip(newslices, dims, range(len(dims))):
                start = s.start if s.start is not None else 0
                stop = s.stop if s.stop is not None else d
                size_arr[i] = stop - start
                offset_arr[i] = start
        except:
            free(size_arr)
            free(offset_arr)
            raise
        
        return (<size_t> size_arr, <size_t> offset_arr)
        
    def _construct_with(self, dims:Tuple[uint64]) -> Tuple[pdcid, Tuple[uint64, ...]]:
        '''
        Constructs a pdc region based on the provided dimensions of the object.
        '''
        size_ptr, offset_ptr = self._get_sizes_offsets(dims)
        cdef uint64_t *size_arr = <uint64_t *> <size_t> size_ptr
        cdef uint64_t *offset_arr = <uint64_t *> <size_t> offset_ptr
        cdef pdcid_t id
        try:
            id = cpdc.PDCregion_create(len(dims), offset_arr, size_arr)
            ctrace('region_create', id, len(dims), tuple(offset_arr[i] for i in range(len(dims))), tuple(size_arr[i] for i in range(len(dims))))
            if id == 0:
                raise PDCError('failed to create region')
            return id, tuple(size_arr[i] for i in range(len(dims)))
        finally:
            _free_from_int(size_ptr)
            _free_from_int(offset_ptr)
        
    def __eq__(self, other:object) -> bool:
        if not isinstance(other, Region):
            return False
        return self.slices == other.slices
    
    def __hash__(self) -> int:
        return hash(tuple(
            (s.start, s.stop) for s in self.slices
        ))
    
    def __repr__(self) -> str:
        return 'region[' + ', '.join(
            f'{"" if s.start is None else s.start}:{"" if s.stop is None else s.stop}' for s in self.slices
        ) + ']'

class _RegionFactory:
    def __getitem__(self, *slices) -> Region:
        return Region(*slices)

region = _RegionFactory()